from . import utils
from .constants import BREAK, POS
from .precedence import OperatorPrecedenceRule


class Mixfix(OperatorPrecedenceRule):
    num_blocks = 2

    def _compile(self, out):
        backtrack = out.var('backtrack', POS)
        with utils.breakable(out):
            with utils.if_succeeds(out, self.operators):
                out += BREAK

            out += POS << backtrack
            self.operand.compile(out)




import ast
import re

from . import expressions as ex
from . import parser
from . import translator


def Grammar(description, name='grammar', include_source=False):
    # Parse the grammar description.
    raw = parser.parse(description)

    # If the grammar is just an expression, create an implicit 'start' rule.
    if not isinstance(raw, list):
        raw = [parser.RuleDef(is_ignored=False, name='start', params=None, expr=raw)]

    # Create the docstring for the module.
    docstring = '# Grammar definition:\n' + description

    # Convert the parse tree into a list of parsing expressions.
    nodes = parser.transform(raw, _create_parsing_expression)

    # Generate and compile the souce code.
    builder = translator.generate_source_code(docstring, nodes)
    return builder.compile(
        module_name=name,
        docstring=docstring,
        source_var='_source_code' if include_source else None,
    )


def _create_parsing_expression(node):
    if isinstance(node, parser.StringLiteral):
        ignore_case = node.value.endswith(('i', 'I'))
        value = ast.literal_eval(node.value[:-1] if ignore_case else node.value)
        if ignore_case:
            return ex.Regex(re.escape(value), ignore_case=True)
        else:
            return ex.Str(value)

    if isinstance(node, parser.RegexLiteral):
        is_binary = node.value.startswith('b')
        ignore_case = node.value.endswith(('i', 'I'))
        value = node.value

        # Remove leading 'b'.
        if is_binary:
            value = value[1:]

        # Remove trailing 'i'.
        if ignore_case:
            value = value[:-1]

        # Remove /slash/ delimiters.
        value = value[1:-1]

        # Enocde binary string.
        if is_binary:
            value = value.encode('ascii')

        return ex.Regex(value, ignore_case=ignore_case)

    if isinstance(node, parser.ByteLiteral):
        return ex.Byte(node.value)

    if isinstance(node, parser.PythonExpression):
        return ex.PythonExpression(node.value)

    if isinstance(node, parser.PythonSection):
        return ex.PythonSection(node.value)

    if isinstance(node, parser.Ref):
        return ex.Ref(node.value)

    if isinstance(node, parser.LetExpression):
        return ex.Let(node.name, node.expr, node.body)

    if isinstance(node, parser.ListLiteral):
        return ex.Seq(*node.elements)

    if isinstance(node, parser.ArgList):
        return node

    if isinstance(node, parser.Postfix) and isinstance(node.operator, parser.ArgList):
        left, args = node.left, node.operator.args
        if isinstance(left, ex.Ref) and hasattr(ex, left.name):
            return getattr(ex, left.name)(
                *[unwrap(x) for x in args if not isinstance(x, ex.KeywordArg)],
                **{x.name: unwrap(x.expr) for x in args if isinstance(x, ex.KeywordArg)},
            )
        else:
            return ex.Call(left, args)

    if isinstance(node, parser.Postfix):
        classes = {
            '?': ex.Opt,
            '*': ex.List,
            '+': ex.Some,
        }
        if isinstance(node.operator, str) and node.operator in classes:
            return classes[node.operator](node.left)

        if isinstance(node.operator, parser.Repeat):
            start = uncook(node.operator.start)
            stop = uncook(node.operator.stop)
            return ex.List(node.left, min_len=start, max_len=stop)

    if isinstance(node, parser.Repeat):
        return node

    if isinstance(node, parser.Infix) and node.operator == '|':
        left, right = node.left, node.right
        left = list(left.exprs) if isinstance(left, ex.Choice) else [left]
        right = list(right.exprs) if isinstance(right, ex.Choice) else [right]
        return ex.Choice(*left, *right)

    if isinstance(node, parser.Infix):
        classes = {
            '|>': lambda a, b: ex.Apply(a, b, apply_left=False),
            '<|': lambda a, b: ex.Apply(a, b, apply_left=True),
            '/?': lambda a, b: ex.Sep(a, b, allow_trailer=True),
            '//': lambda a, b: ex.Sep(a, b, allow_trailer=False),
            '<<': ex.Left,
            '>>': ex.Right,
            'where': ex.Where,
        }
        return classes[node.operator](node.left, node.right)

    if isinstance(node, parser.KeywordArg):
        return ex.KeywordArg(node.name, node.expr)

    if isinstance(node, parser.RuleDef):
        return ex.Rule(node.name, node.params, node.expr, is_ignored=node.is_ignored)

    if isinstance(node, parser.ClassDef):
        return ex.Class(node.name, node.params, node.members)

    if isinstance(node, parser.ClassMember):
        return ex.Rule(node.name, node.params, node.expr, is_omitted=node.is_omitted)

    if isinstance(node, parser.IgnoreStmt):
        return ex.Rule(None, None, node.expr, is_ignored=True)

    # Otherwise, fail if we don't know what to do with this node.
    raise Exception(f'Unexpected expression: {node!r}')


def unwrap(x):
    return eval(x.source_code) if isinstance(x, ex.PythonExpression) else x


def uncook(x):
    if x is None:
        return None
    if isinstance(x, ex.PythonExpression) and x.source_code == 'None':
        return None
    if isinstance(x, ex.PythonExpression):
        return x.source_code
    if isinstance(x, ex.Ref):
        return x.name

    raise Exception(f'Expected name or Python expression. Received: {x}')



from outsourcer import Code

from . import utils
from .base import Expression
from .constants import BREAK, POS, RESULT


class Seq(Expression):
    is_commented = False
    num_blocks = 2

    def __init__(self, *exprs, names=None, constructor=None, constructor_args=None):
        if isinstance(constructor, type):
            constructor = constructor.__name__
        self.exprs = exprs

        if names is not None:
            if len(names) != len(exprs):
                raise Exception('Expected same number of expressions and names.')
            self.names = names
        else:
            self.names = [None] * len(exprs)

        self.needs_parse_info = constructor is not None
        self.constructor = None if constructor is None else Code(constructor)
        self.constructor_args = constructor_args

    def __str__(self):
        return f'[{", ".join(str(x) for x in self.exprs)}]'

    def _compile(self, out):
        if self.needs_parse_info:
            start_pos = out.var('start_pos', POS)

        cargs = self.constructor_args
        which = None if cargs is None else set(cargs)

        with utils.breakable(out):
            items = []
            for name, expr in zip(self.names, self.exprs):
                with utils.if_fails(out, expr):
                    out += BREAK

                item = out.var('item') if name is None else Code(name)
                out += item << RESULT

                if which is None or name in which:
                    items.append(item)

            result = items if self.constructor is None else self.constructor(*items)
            out += RESULT << result

            if self.needs_parse_info:
                out += RESULT._metadata.position_info << (start_pos, POS)



from . import utils
from .base import Expression
from .constants import BREAK, POS, RESULT
from .fail import Fail


class Choice(Expression):
    is_commented = False
    num_blocks = 2

    def __init__(self, *exprs):
        self.exprs = exprs

    def __str__(self):
        return ' | '.join(str(x) for x in self.exprs)

    def operand_string(self):
        return f'({self})'

    def always_succeeds(self):
        return any(x.always_succeeds() for x in self.exprs)

    def can_partially_succeed(self):
        return not self.always_succeeds() and (
            any(x.can_partially_succeed() for x in self.exprs)
        )

    def _compile(self, out):
        needs_err = not self.always_succeeds()
        needs_backtrack = any(x.can_partially_succeed() for x in self.exprs)

        backtrack = out.var('backtrack') if needs_backtrack else None
        farthest_pos = out.var('farthest_pos') if needs_err else None

        if needs_err:
            farthest_err = out.var('farthest_err', self.error_func())

        if needs_err and needs_backtrack:
            out += backtrack << farthest_pos << POS
        elif needs_backtrack:
            out += backtrack << POS
        elif needs_err:
            out += farthest_pos << POS

        with utils.breakable(out):
            for i, expr in enumerate(self.exprs):
                comment = f'Option {i+1}:'
                if expr.always_succeeds():
                    comment += ' (always_succeeds)'
                out.add_comment(comment)

                with utils.if_succeeds(out, expr):
                    if expr.always_succeeds():
                        break
                    else:
                        out += BREAK

                if needs_err and expr.can_partially_succeed():
                    if isinstance(expr, Fail):
                        condition = farthest_pos <= POS
                    else:
                        condition = farthest_pos < POS

                    with out.IF(condition):
                        out += farthest_pos << POS
                        out += farthest_err << RESULT

                if i + 1 < len(self.exprs) and expr.can_partially_succeed():
                    out += POS << backtrack

            if needs_err:
                out += POS << farthest_pos
                out += RESULT << farthest_err

    def complain(self):
        return 'Unexpected input'



from outsourcer import Code

from . import utils
from .base import Expression
from .constants import POS, RESULT, STATUS


class Skip(Expression):
    num_blocks = 2

    def __init__(self, *exprs):
        self.exprs = exprs

    def __str__(self):
        return f'Skip({", ".join(str(x) for x in self.exprs)})'

    def always_succeeds(self):
        return True

    def can_partially_succeed(self):
        return False

    def _compile(self, out):
        checkpoint = out.var('checkpoint')

        with utils.breakable(out):
            out += checkpoint << POS

            for expr in self.exprs:
                expr.compile(out)

                if expr.always_succeeds():
                    with out.IF(POS != checkpoint):
                        out += Code('continue')
                    continue

                with out.IF(STATUS):
                    out += Code('continue')

                if expr.can_partially_succeed():
                    with out.ELSE():
                        out += POS << checkpoint

        out += RESULT << None
        out += STATUS << True



from outsourcer import Code

from . import utils
from .constants import BREAK, POS, RESULT, STATUS
from .precedence import OperatorPrecedenceRule


class Prefix(OperatorPrecedenceRule):
    num_blocks = 2

    def _compile(self, out):
        prev = out.var('prev', None)
        checkpoint = out.var('checkpoint', POS)
        staging = out.var('staging')

        with out.WHILE(True):
            with utils.if_fails(out, self.operators):
                out += POS << checkpoint
                out += BREAK

            out += checkpoint << POS
            step = out.var('step', Code('Prefix')(RESULT, None))

            with out.IF(Code(prev, ' is ', None)):
                out += prev << staging << step

            with out.ELSE():
                out += prev.right << step
                out += prev << step

        self.operand.compile(out)

        with out.IF(Code(STATUS, ' and ', prev)):
            out += prev.right << RESULT
            out += RESULT << staging



from outsourcer import Code

from . import utils
from .constants import BREAK, POS, RESULT, STATUS
from .precedence import OperatorPrecedenceRule


class Postfix(OperatorPrecedenceRule):
    num_blocks = 3

    def _compile(self, out):
        with utils.if_succeeds(out, self.operand):
            staging = out.var('staging', RESULT)
            checkpoint = out.var('checkpoint', POS)

            with out.WHILE(True):
                self.operators.compile(out)

                with out.IF(STATUS):
                    out += staging << Code('Postfix')(staging, RESULT)
                    out += checkpoint << POS

                with out.ELSE():
                    out += POS << checkpoint
                    out += RESULT << staging
                    out += STATUS << True
                    out += BREAK



from outsourcer import Code

from . import utils
from .base import Expression
from .constants import POS, RESULT, STATUS, TEXT


class Rule(Expression):
    has_params = True

    is_tagged = False
    is_commented = False

    num_blocks = 1

    def __init__(self, name, params, expr, is_ignored=False, is_omitted=False):
        self.name = name
        self.params = params
        self.expr = expr
        self.is_ignored = is_ignored
        self.is_omitted = is_omitted

    def __str__(self):
        params = '' if self.params is None else f'({", ".join(self.params)})'
        return f'{self.name}{params} = {self.expr}'

    def _compile(self, out):
        params = [str(TEXT), str(POS)] + (self.params or [])
        impl_name = utils.implementation_name(self.name)
        entry_name = f'_parse_{self.name}'

        definition = str(self)
        if '"""' in definition:
            definition = definition.replace('"""', '\\"\\"\\"')

        with out.global_section():
            with out.DEF(impl_name, params):
                out.add_comment(f'Rule {self.name!r}')
                self.expr.compile(out)
                out.YIELD((STATUS, RESULT, POS))

            with out.DEF(entry_name, ['text', 'pos=0', 'fullparse=True']):
                out.RETURN(Code(f'_run(text, pos, {impl_name}, fullparse)'))

            out += Code(f'{self.name} = Rule({self.name!r}, {entry_name}, """')
            out.extend(Code('    ', x) for x in definition.split('\n'))
            out += Code('""")')



from outsourcer import Code

from . import utils
from .base import Expression
from .constants import BREAK, POS, RESULT, STATUS


class List(Expression):
    num_blocks = 2

    def __init__(self, expr, min_len=None, max_len=None):
        self.expr = expr
        self.min_len = min_len
        self.max_len = max_len
        _check_min_and_max_len(min_len, max_len)

    def __str__(self):
        arg = self.expr.operand_string()

        if self.min_len is None and self.max_len is None:
            op = '*'
        elif self.min_len == 1 and self.max_len is None:
            op = '+'
        elif self.min_len == self.max_len:
            op = f'{{{self.min_len}}}'
        else:
            op = f'{{{self.min_len or 0},{self.max_len}}}'

        return f'{arg}{op}'

    def always_succeeds(self):
        return not self.min_len or self.min_len == '0'

    def can_partially_succeed(self):
        return not self.always_succeeds() and self.expr.can_partially_succeed()

    def _compile(self, out):
        if self.max_len == 0 or self.max_len == '0':
            out += RESULT << []
            out += STATUS << True
            return

        LEN = Code('len')
        staging = out.var('staging', [])

        with out.WHILE(True):
            if self.expr.can_partially_succeed():
                checkpoint = out.var('checkpoint', POS)

            with utils.if_fails(out, self.expr):
                if self.expr.can_partially_succeed():
                    out += POS << checkpoint
                out += BREAK

            out += staging.append(RESULT)

            if self.max_len is not None:
                with out.IF(LEN(staging) == Code(self.max_len)):
                    out += BREAK

        if not self.min_len or self.min_len == '0':
            out += RESULT << staging
            out += STATUS << True
            return

        if self.min_len == 1 or self.min_len == '1':
            condition = staging
        else:
            condition = LEN(staging) >= Code(self.min_len)

        with out.IF(condition):
            out += RESULT << staging
            out += STATUS << True


def _check_min_and_max_len(min_len, max_len):
    if min_len is None or max_len is None:
        return

    try:
        min_len, max_len = int(min_len), int(max_len)
    except ValueError:
        return

    if min_len > max_len:
        raise Exception('Expected min_len to be less than max_len')



from contextlib import contextmanager
from outsourcer import Code, Yield
from .constants import BREAK, CALL, STATUS


@contextmanager
def if_succeeds(out, expr):
    expr.compile(out)
    if expr.always_succeeds():
        yield
    else:
        with out.IF(STATUS):
            yield


@contextmanager
def if_fails(out, expr):
    expr.compile(out)
    if expr.always_succeeds():
        with out._sandbox():
            yield
    else:
        with out.IF_NOT(STATUS):
            yield


@contextmanager
def breakable(out):
    with out.WHILE(True):
        yield
        out += BREAK


def infix_str(expr1, op, expr2):
    arg1 = expr1.operand_string()
    arg2 = expr2.operand_string()
    return f'{arg1} {op} {arg2}'


def skip_ignored(pos):
    return Yield((CALL, Code(implementation_name('_ignored')), pos))[2]


def implementation_name(name):
    return f'_try_{name}'
